# Copyright 2021-2023 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Helpful utilities that can be used together with IQMClient.
"""
from json import JSONEncoder, dumps, loads
from typing import Any, TypeVar

import numpy as np

T = TypeVar('T')


class IQMJSONEncoder(JSONEncoder):
    """JSONEncoder that that adds support for some non-JSON datatypes"""

    def default(self, o: Any):
        if isinstance(o, np.ndarray):
            return o.tolist()
        return JSONEncoder.default(self, o)


def to_json_dict(obj: dict[str, Any]) -> dict:
    """Convert a dict to JSON serializable dict

    Args:
        obj: dict to convert

    Returns:
        dict containing converted data

    Raises:
        ValueError if the original dict contains unsupported datatypes"""
    try:
        return loads(dumps(obj, allow_nan=False, cls=IQMJSONEncoder))
    except (ValueError, TypeError) as e:
        raise ValueError('Object contains values that are not JSON serializable') from e



# Configuration file for the Sphinx documentation builder.
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

# Find the path to the source files we want to to document, relative to the location of this file,
# convert it to an absolute path.
py_path = os.path.join(os.getcwd(), os.path.dirname(__file__), '../src')
sys.path.insert(0, os.path.abspath(py_path))

# -- Project information -----------------------------------------------------

project = 'IQM client'
copyright = '2021-2024, IQM'
author = 'IQM client developers'

# The short X.Y version.
version = ''
# The full version, including alpha/beta/rc tags.
release = ''
try:
    from iqm.iqm_client import __version__ as version
except ImportError:
    pass
else:
    release = version

# -- General configuration ---------------------------------------------------

# require a recent version of Sphinx
needs_sphinx = '7.2'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx.ext.mathjax',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.todo',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# Include extra files in the HTML docs.
html_extra_path = []

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '.*']

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = '%Y-%m-%d'

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = True


# -- Autodoc ------------------------------------------------------------

# member ordering in autodoc output (default: 'alphabetical')
autodoc_member_order = 'bysource'

# where should signature annotations appear in the docs, function signature or parameter description?
autodoc_typehints = 'description'
# autodoc_typehints = 'description' puts the __init__ annotations into its docstring,
# which we thus have to include in the class documentation.
autoclass_content = 'class'

# Sphinx 3.3+: manually clean up type alias rendering in the docs
# autodoc_type_aliases = {'TypeAlias': 'exa.experiment.somemodule.TypeAlias'}


# -- Autosummary ------------------------------------------------------------

# use autosummary to generate stub pages for API docs
autosummary_generate = True


# -- Options for HTML output -------------------------------------------------

import sphinx_book_theme

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_book_theme'

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [sphinx_book_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {}

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '_static/images/logo.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = '_static/images/favicon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%Y-%m-%d'

# Output file base name for HTML help builder.
htmlhelp_basename = 'iqm_client-doc'


# -- MathJax options ----------------------------------------------------------

# Here we configure MathJax, mostly to define LaTeX macros.
mathjax3_config = {
    'tex': {
        'macros': {
            'vr': r'\vec{r}',  # no arguments
            'ket': [r'\left| #1 \right\rangle', 1],  # one argument
            'iprod': [r'\left\langle #1 | #2 \right\rangle', 2],  # two arguments
        }
    }
}


# -- External mapping ------------------------------------------------------------

python_version = '.'.join(map(str, sys.version_info[0:2]))
intersphinx_mapping = {
    'python': ('https://docs.python.org/' + python_version, None),
}

extlinks = {
    'issue': ('https://github.com/iqm-finland/iqm-client/issues/%s', 'issue '),
    'mr': ('https://github.com/iqm-finland/iqm-client/pull/%s', 'MR '),
}


# -- Options for sphinxcontrib.bibtex -------------------------------------------------

# List of all bibliography files used.
#bibtex_bibfiles = ['references.bib']



# Copyright 2021-2022 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Mocks server calls for testing
"""

from base64 import b64encode
import json
import os
import platform
import time
from typing import Any, Optional
from uuid import UUID

from mockito import expect, mock, when
from mockito.matchers import ANY
import pytest
import requests
from requests import HTTPError, Response

from iqm.iqm_client import (
    AUTH_CLIENT_ID,
    AUTH_REALM,
    DIST_NAME,
    REQUESTS_TIMEOUT,
    AuthRequest,
    Circuit,
    GrantType,
    HeraldingMode,
    Instruction,
    IQMClient,
    RunRequest,
    SingleQubitMapping,
    __version__,
)


@pytest.fixture()
def base_url() -> str:
    return 'https://example.com'


@pytest.fixture()
def client_signature() -> str:
    return 'some-signature'


@pytest.fixture()
def sample_calibration_set_id() -> UUID:
    return UUID('9ddb9586-8f27-49a9-90ed-41086b47f6bd')


@pytest.fixture()
def existing_run_id() -> UUID:
    return UUID('3c3fcda3-e860-46bf-92a4-bcc59fa76ce9')


@pytest.fixture()
def missing_run_id() -> UUID:
    return UUID('059e4186-50a3-4e6c-ba1f-37fe6afbdfc2')


@pytest.fixture()
def sample_client(base_url) -> IQMClient:
    return IQMClient(url=base_url)


@pytest.fixture()
def client_with_signature(base_url) -> IQMClient:
    return IQMClient(url=base_url, client_signature='some-signature')


@pytest.fixture()
def credentials(base_url):
    return {
        'auth_server_url': f'{base_url}/auth',
        'username': 'some_username',
        'password': 'some_password',
    }


@pytest.fixture()
def client_with_credentials(base_url, credentials):
    prepare_tokens(300, 300, **credentials)
    return IQMClient(url=base_url, **credentials)


@pytest.fixture()
def client_with_external_token(tokens_dict, credentials):
    prepare_tokens(300, 300, **credentials)
    tokens_path = os.path.dirname(os.path.realpath(__file__)) + '/resources/tokens.json'
    return IQMClient(tokens_dict['auth_server_url'], tokens_file=tokens_path)


@pytest.fixture()
def jobs_url(base_url) -> str:
    return f'{base_url}/jobs'


@pytest.fixture()
def existing_job_url(jobs_url, existing_run_id) -> str:
    return f'{jobs_url}/{existing_run_id}'


@pytest.fixture()
def existing_job_status_url(existing_job_url) -> str:
    return f'{existing_job_url}/status'


@pytest.fixture()
def quantum_architecture_url(base_url) -> str:
    return f'{base_url}/quantum-architecture'


@pytest.fixture
def settings_dict():
    """
    Reads and parses settings file into a dictionary
    """
    settings_path = os.path.dirname(os.path.realpath(__file__)) + '/resources/settings.json'
    with open(settings_path, 'r', encoding='utf-8') as f:
        return json.loads(f.read())


@pytest.fixture
def tokens_dict():
    """
    Reads and parses tokens file into a dictionary
    """
    tokens_path = os.path.dirname(os.path.realpath(__file__)) + '/resources/tokens.json'
    with open(tokens_path, 'r', encoding='utf-8') as f:
        return json.loads(f.read())


@pytest.fixture()
def sample_circuit_metadata():
    return {'experiment_type': 'test', 'qubits': (0, 1), 'values': [0.01686514, 0.05760602]}


@pytest.fixture
def sample_circuit(sample_circuit_metadata):
    """
    A sample circuit for testing submit_circuit
    """
    return create_sample_circuit(['QB1', 'QB2'], metadata=sample_circuit_metadata)


@pytest.fixture
def sample_circuit_logical(sample_circuit_metadata):
    """
    A sample circuit with logical names for testing submit_circuit
    """
    return create_sample_circuit(['Qubit A', 'Qubit B'], metadata=sample_circuit_metadata)


def create_sample_circuit(qubits: list[str], metadata) -> Circuit:
    return Circuit(
        name='The circuit',
        instructions=[
            Instruction(
                name='cz',
                qubits=tuple(qubits),
                args={},
            ),
            Instruction(
                name='phased_rx',
                implementation='drag_gaussian',
                qubits=(qubits[0],),
                args={'phase_t': 0.7, 'angle_t': 0.25},
            ),
            Instruction(
                name='prx',
                qubits=(qubits[0],),
                args={'phase_t': 0.3, 'angle_t': -0.2},
            ),
            Instruction(name='measurement', qubits=(qubits[0],), args={'key': 'A'}),
            Instruction(name='measure', qubits=(qubits[1],), args={'key': 'B'}),
        ],
        metadata=metadata,
    )


@pytest.fixture
def sample_circuit_with_raw_instructions(sample_circuit_metadata):
    """
    A sample circuit with instructions defined by dicts for testing if
    we do not break pydantic parsing logic with custom validators
    """
    return Circuit(
        name='The circuit',
        instructions=[
            {
                'name': 'cz',
                'qubits': (
                    'Qubit A',
                    'Qubit B',
                ),
                'args': {},
            },
            {
                'name': 'phased_rx',
                'implementation': 'drag_gaussian',
                'qubits': ('Qubit A',),
                'args': {'phase_t': 0.7, 'angle_t': 0.25},
            },
            {
                'name': 'prx',
                'qubits': ('Qubit A',),
                'args': {'phase_t': 0.3, 'angle_t': -0.2},
            },
            {'name': 'measurement', 'qubits': ('Qubit A',), 'args': {'key': 'A'}},
            {'name': 'measure', 'qubits': ('Qubit B',), 'args': {'key': 'B'}},
        ],
        metadata=sample_circuit_metadata,
    )


@pytest.fixture()
def minimal_run_request(sample_circuit) -> RunRequest:
    return RunRequest(
        circuits=[sample_circuit],
        shots=10,
    )


@pytest.fixture()
def run_request_with_heralding(sample_circuit) -> RunRequest:
    return RunRequest(
        circuits=[sample_circuit],
        shots=10,
        heralding_mode=HeraldingMode.ZEROS,
    )


@pytest.fixture()
def run_request_with_custom_settings(sample_circuit_logical, settings_dict) -> RunRequest:
    return RunRequest(
        circuits=[sample_circuit_logical],
        shots=10,
        qubit_mapping=[
            SingleQubitMapping(logical_name='Qubit A', physical_name='QB1'),
            SingleQubitMapping(logical_name='Qubit B', physical_name='QB2'),
        ],
        custom_settings=settings_dict,
        heralding_mode=HeraldingMode.NONE,
    )


@pytest.fixture()
def run_request_without_qubit_mapping(sample_circuit) -> RunRequest:
    return RunRequest(
        circuits=[sample_circuit],
        shots=10,
        heralding_mode=HeraldingMode.NONE,
    )


@pytest.fixture()
def run_request_with_invalid_qubit_mapping(sample_circuit_logical) -> RunRequest:
    return RunRequest(
        circuits=[sample_circuit_logical],
        shots=10,
        qubit_mapping=[
            SingleQubitMapping(logical_name='Qubit A', physical_name='QB1'),
            SingleQubitMapping(logical_name='Qubit B', physical_name='QB1'),
        ],
        heralding_mode=HeraldingMode.NONE,
    )


@pytest.fixture()
def run_request_with_incomplete_qubit_mapping(sample_circuit_logical) -> RunRequest:
    return RunRequest(
        circuits=[sample_circuit_logical],
        shots=10,
        qubit_mapping=[
            SingleQubitMapping(logical_name='Qubit A', physical_name='QB1'),
        ],
        heralding_mode=HeraldingMode.NONE,
    )


@pytest.fixture()
def run_request_with_calibration_set_id(sample_circuit_logical, sample_calibration_set_id) -> RunRequest:
    return RunRequest(
        circuits=[sample_circuit_logical],
        shots=10,
        qubit_mapping=[
            SingleQubitMapping(logical_name='Qubit A', physical_name='QB1'),
            SingleQubitMapping(logical_name='Qubit B', physical_name='QB2'),
        ],
        calibration_set_id=sample_calibration_set_id,
        heralding_mode=HeraldingMode.NONE,
    )


@pytest.fixture()
def run_request_with_duration_check_disabled(sample_circuit_logical) -> RunRequest:
    return RunRequest(
        circuits=[sample_circuit_logical],
        shots=10,
        qubit_mapping=[
            SingleQubitMapping(logical_name='Qubit A', physical_name='QB1'),
            SingleQubitMapping(logical_name='Qubit B', physical_name='QB2'),
        ],
        max_circuit_duration_over_t2=0.0,
        heralding_mode=HeraldingMode.NONE,
    )


@pytest.fixture()
def pending_compilation_job_result(sample_circuit):
    return MockJsonResponse(
        200, {'status': 'pending compilation', 'metadata': {'request': {'shots': 10, 'circuits': [sample_circuit]}}}
    )


@pytest.fixture()
def pending_execution_job_result(sample_circuit):
    return MockJsonResponse(
        200, {'status': 'pending execution', 'metadata': {'request': {'shots': 10, 'circuits': [sample_circuit]}}}
    )


@pytest.fixture()
def pending_deletion_job_result(sample_circuit):
    return MockJsonResponse(
        200, {'status': 'pending deletion', 'metadata': {'request': {'shots': 10, 'circuits': [sample_circuit]}}}
    )


@pytest.fixture()
def deleted_job_result():
    return MockJsonResponse(200, {'status': 'deleted', 'metadata': {'request': {'shots': 1, 'circuits': []}}})


@pytest.fixture()
def ready_job_result(sample_circuit, sample_calibration_set_id):
    return MockJsonResponse(
        200,
        {
            'status': 'ready',
            'measurements': [{'result': [[1, 0, 1, 1], [1, 0, 0, 1], [1, 0, 1, 1], [1, 0, 1, 1]]}],
            'metadata': {
                'calibration_set_id': str(sample_calibration_set_id),
                'request': {
                    'shots': 42,
                    'circuits': [sample_circuit],
                    'calibration_set_id': str(sample_calibration_set_id),
                },
                'timestamps': {
                    'job_start': '0.0',
                    'job_end': '1.1',
                },
            },
        },
    )


@pytest.fixture()
def job_result_with_warnings(sample_circuit, sample_calibration_set_id):
    return MockJsonResponse(
        200,
        {
            'status': 'ready',
            'metadata': {
                'calibration_set_id': str(sample_calibration_set_id),
                'request': {
                    'shots': 42,
                    'circuits': [sample_circuit],
                    'calibration_set_id': str(sample_calibration_set_id),
                },
                'timestamps': {
                    'job_start': '0.0',
                    'job_end': '1.1',
                },
            },
            'warnings': ['This is a warning message'],
        },
    )


@pytest.fixture()
def pending_compilation_status():
    return MockJsonResponse(200, {'status': 'pending compilation'})


@pytest.fixture()
def pending_execution_status():
    return MockJsonResponse(200, {'status': 'pending execution'})


@pytest.fixture()
def ready_status():
    return MockJsonResponse(200, {'status': 'ready'})


@pytest.fixture()
def pending_deletion_status():
    return MockJsonResponse(200, {'status': 'pending deletion'})


@pytest.fixture()
def deleted_status():
    return MockJsonResponse(200, {'status': 'deleted'})


@pytest.fixture
def sample_quantum_architecture():
    return {
        'quantum_architecture': {
            'name': 'hercules',
            'qubits': ['QB1', 'QB2'],
            'qubit_connectivity': [['QB1', 'QB2']],
            'operations': {
                'phased_rx': [['QB1'], ['QB2']],
                'cz': [['QB1', 'QB2']],
                'measurement': [['QB1'], ['QB2']],
            },
        }
    }


@pytest.fixture
def sample_move_architecture():
    return {
        'quantum_architecture': {
            'name': 'hercules',
            'qubits': ['COMP_R', 'QB1', 'QB2', 'QB3'],
            'qubit_connectivity': [
                ['QB1', 'COMP_R'],
                ['QB2', 'COMP_R'],
                ['QB3', 'COMP_R'],
            ],
            'operations': {
                'prx': [['QB1'], ['QB2'], ['QB3']],
                'cz': [['QB1', 'COMP_R'], ['QB2', 'COMP_R']],
                'move': [['QB3', 'COMP_R']],
                'measure': [['QB1'], ['QB2'], ['QB3']],
                'barrier': [],
            },
        }
    }


class MockTextResponse:
    def __init__(self, status_code: int, text: str, history: Optional[list[Response]] = None):
        self.status_code = status_code
        self.text = text
        self.history = history

    def json(self):
        return json.loads(self.text)

    def raise_for_status(self):
        if 400 <= self.status_code < 600:
            raise HTTPError('')


@pytest.fixture()
def not_valid_json_response() -> MockTextResponse:
    return MockTextResponse(200, 'not a valid json')


class MockJsonResponse:
    def __init__(self, status_code: int, json_data: dict, history: Optional[list[Response]] = None):
        self.status_code = status_code
        self.json_data = json_data
        self.history = history

    @property
    def text(self):
        return json.dumps(self.json_data)

    def json(self):
        return self.json_data

    def raise_for_status(self):
        if 400 <= self.status_code < 600:
            raise HTTPError(f'{self.status_code}', response=self)


@pytest.fixture()
def not_valid_client_configuration_response() -> MockJsonResponse:
    return MockJsonResponse(400, {'detail': 'not a valid client configuration'})


@pytest.fixture()
def submit_success(existing_run_id) -> MockJsonResponse:
    return MockJsonResponse(201, {'id': str(existing_run_id)})


@pytest.fixture()
def submit_failed_auth() -> MockJsonResponse:
    return MockJsonResponse(401, {'detail': 'unauthorized'})


@pytest.fixture()
def quantum_architecture_success(sample_quantum_architecture) -> MockJsonResponse:
    return MockJsonResponse(200, sample_quantum_architecture)


@pytest.fixture()
def abort_job_success() -> MockJsonResponse:
    return MockJsonResponse(200, {})


@pytest.fixture()
def abort_job_failed() -> MockJsonResponse:
    return MockJsonResponse(400, {'detail': 'failed to abort job'})


def prepare_tokens(
    access_token_lifetime: int,
    refresh_token_lifetime: int,
    previous_refresh_token: Optional[str] = None,
    status_code: int = 200,
    **credentials,
) -> dict[str, str]:
    """Prepare tokens and set them to be returned for a token request.

    Args:
        access_token_lifetime: seconds from current time to access token expire time
        refresh_token_lifetime: seconds from current time to refresh token expire time
        previous_refresh_token: refresh token to be used in refresh request
        status_code: status code to return for token request
        credentials: dict containing auth_server_url, username and password

    Returns:
         Prepared tokens as a dict.
    """
    if previous_refresh_token is None:
        request_data = AuthRequest(
            client_id=AUTH_CLIENT_ID,
            grant_type=GrantType.PASSWORD,
            username=credentials['username'],
            password=credentials['password'],
        )
    else:
        request_data = AuthRequest(
            client_id=AUTH_CLIENT_ID, grant_type=GrantType.REFRESH, refresh_token=previous_refresh_token
        )

    tokens = {
        'access_token': make_token('Bearer', access_token_lifetime),
        'refresh_token': make_token('Refresh', refresh_token_lifetime),
    }
    when(requests).post(
        f'{credentials["auth_server_url"]}/realms/{AUTH_REALM}/protocol/openid-connect/token',
        data=request_data.model_dump(exclude_none=True),
        timeout=REQUESTS_TIMEOUT,
    ).thenReturn(MockJsonResponse(status_code, tokens))

    return tokens


def make_token(token_type: str, lifetime: int) -> str:
    """Encode given token type and expire time as a token.

    Args:
        token_type: 'Bearer' for access tokens, 'Refresh' for refresh tokens
        lifetime: seconds from current time to token's expire time

    Returns:
        Encoded token
    """
    empty = b64encode('{}'.encode('utf-8')).decode('utf-8')
    body = f'{{ "typ": "{token_type}", "exp": {int(time.time()) + lifetime} }}'
    body = b64encode(body.encode('utf-8')).decode('utf-8')
    return f'{empty}.{body}.{empty}'


def expect_logout(auth_server_url: str, refresh_token: str, timeout: float = ANY(float)):
    """Prepare for logout request.

    Args:
        auth_server_url: base URL of the authentication server
        refresh_token: refresh token expected to be used in the request
        timeout: expected timeout value, if any
    """
    request_data = AuthRequest(client_id=AUTH_CLIENT_ID, refresh_token=refresh_token)
    expect(requests, times=1).post(
        f'{auth_server_url}/realms/{AUTH_REALM}/protocol/openid-connect/logout',
        data=request_data.model_dump(exclude_none=True),
        timeout=timeout,
    ).thenReturn(mock({'status_code': 204, 'text': '{}'}))


def post_jobs_args(
    run_request: Optional[RunRequest] = None,
    user_agent: Optional[str] = None,
    access_token: Optional[str] = None,
) -> dict[str, Any]:
    """Returns expected kwargs of POST /jobs request"""
    headers = {'Expect': '100-Continue'} if run_request is not None else {}
    signature = f'{platform.platform(terse=True)}, python {platform.python_version()}, {DIST_NAME} {__version__}'
    headers['User-Agent'] = signature if user_agent is None else user_agent
    if access_token is not None:
        headers['Authorization'] = f'Bearer {access_token}'
    if run_request is None:
        return {'headers': headers, 'timeout': REQUESTS_TIMEOUT}
    return {
        'json': json.loads(run_request.model_dump_json(exclude_none=True)),
        'headers': headers,
        'timeout': REQUESTS_TIMEOUT,
    }


def get_jobs_args(
    user_agent: Optional[str] = None,
    access_token: Optional[str] = None,
) -> dict[str, Any]:
    """Returns expected kwargs of POST /jobs request"""
    headers = {}
    signature = f'{platform.platform(terse=True)}, python {platform.python_version()}, {DIST_NAME} {__version__}'
    headers['User-Agent'] = signature if user_agent is None else user_agent
    if access_token is not None:
        headers['Authorization'] = f'Bearer {access_token}'
    return {
        'headers': headers if headers else None,
        'timeout': REQUESTS_TIMEOUT,
    }


def submit_circuits_args(run_request: RunRequest) -> dict[str, Any]:
    """Return args to be used with submit_circuits to generate the expected RunRequest"""
    qm_dict = None
    if run_request.qubit_mapping is not None:
        qm_dict = {qm.logical_name: qm.physical_name for qm in run_request.qubit_mapping}
    return {
        'circuits': run_request.circuits,
        'qubit_mapping': qm_dict,
        'custom_settings': run_request.custom_settings,
        'calibration_set_id': run_request.calibration_set_id,
        'shots': run_request.shots,
        'max_circuit_duration_over_t2': run_request.max_circuit_duration_over_t2,
        'heralding_mode': run_request.heralding_mode,
    }



# Copyright 2024 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
This module contains the definition of the quantum architecture specification
classes.

For more information on the topic, refer to the more comprehensive documentation in iqm_client.py.
"""
from __future__ import annotations

from pydantic import BaseModel, Field

from iqm.iqm_client.instruction import get_current_instruction_name, is_directed_instruction, is_multi_qubit_instruction


class QuantumArchitectureSpecification(BaseModel):
    """Quantum architecture specification."""

    name: str = Field(...)
    """Name of the quantum architecture."""
    operations: dict[str, list[list[str]]] = Field(...)
    """Operations supported by this quantum architecture, mapped to the allowed loci."""
    qubits: list[str] = Field(...)
    """List of qubits of this quantum architecture."""
    qubit_connectivity: list[list[str]] = Field(...)
    """Qubit connectivity of this quantum architecture."""

    def __init__(self, **data):
        # Convert a simplified quantum architecture to full quantum architecture
        raw_operations = data.get("operations")
        raw_qubits = data.get("qubits")
        raw_qubit_connectivity = data.get("qubit_connectivity")
        if isinstance(raw_operations, list):
            data["operations"] = {
                get_current_instruction_name(op): raw_qubit_connectivity
                if is_multi_qubit_instruction(get_current_instruction_name(op))
                else [[qb] for qb in raw_qubits]
                for op in raw_operations
            }

        super().__init__(**data)
        self.operations = {get_current_instruction_name(k): v for k, v in self.operations.items()}

    def has_equivalent_operations(self, other: QuantumArchitectureSpecification):
        """Compares the given operation sets defined by the quantum architecture against
        another architecture specification.

        Returns:
             True if the operation and the loci are equivalent.
        """
        return QuantumArchitectureSpecification.compare_operations(self.operations, other.operations)

    @staticmethod
    def compare_operations(ops1: dict[str, list[list[str]]], ops2: dict[str, list[list[str]]]) -> bool:
        """Compares the given operation sets.

        Returns:
             True if the operation and the loci are equivalent.
        """
        if set(ops1.keys()) != set(ops2.keys()):
            return False
        for [op, c1] in ops1.items():
            c2 = ops2[op]
            if is_multi_qubit_instruction(op):
                if not is_directed_instruction(op):
                    c1 = [sorted(qbs) for qbs in c1]
                    c2 = [sorted(qbs) for qbs in c2]
                if sorted(c1) != sorted(c2):
                    return False
            else:
                qs1 = {q for [q] in c1}
                qs2 = {q for [q] in c2}
                if qs1 != qs2:
                    return False
        return True


class QuantumArchitecture(BaseModel):
    """Quantum architecture as returned by server."""

    quantum_architecture: QuantumArchitectureSpecification = Field(...)
    """Details about the quantum architecture."""



import re

import pytest

from iqm.iqm_client import CircuitExecutionError, Instruction, IQMClient, QuantumArchitecture

sample_qb_mapping = {'0': 'COMP_R', '1': 'QB1', '2': 'QB2', '3': 'QB3'}
reverse_qb_mapping = {value: key for key, value in sample_qb_mapping.items()}


@pytest.mark.parametrize('qubits', [['QB1', 'COMP_R'], ['COMP_R', 'QB1'], ['COMP_R', 'QB2']])
def test_allowed_cz_qubits(sample_move_architecture, qubits):
    """
    Tests that instruction validation passes for allowed CZ loci
    """
    arch = QuantumArchitecture(**sample_move_architecture).quantum_architecture
    IQMClient._validate_instruction(arch, Instruction(name='cz', qubits=qubits, args={}), None)
    IQMClient._validate_instruction(
        arch, Instruction(name='cz', qubits=[reverse_qb_mapping[q] for q in qubits], args={}), sample_qb_mapping
    )


@pytest.mark.parametrize(
    'qubits', [['QB1', 'QB2'], ['QB2', 'QB1'], ['QB1', 'QB1'], ['QB3', 'QB1'], ['COMP_R', 'COMP_R'], ['COMP_R', 'QB3']]
)
def test_disallowed_cz_qubits(sample_move_architecture, qubits):
    """
    Tests that instruction validation fails for loci that are not allowed for CZ by the quantum architecture
    """
    arch = QuantumArchitecture(**sample_move_architecture).quantum_architecture
    with pytest.raises(CircuitExecutionError) as err:
        IQMClient._validate_instruction(arch, Instruction(name='cz', qubits=qubits, args={}), None)
    assert str(err.value) == f"('{qubits[0]}', '{qubits[1]}') not allowed as locus for cz"

    reversed_qb = [reverse_qb_mapping[q] for q in qubits]
    with pytest.raises(CircuitExecutionError) as err2:
        IQMClient._validate_instruction(arch, Instruction(name='cz', qubits=reversed_qb, args={}), sample_qb_mapping)
    assert (
        str(err2.value)
        == f"('{reversed_qb[0]}', '{reversed_qb[1]}') = ('{qubits[0]}', '{qubits[1]}') not allowed as locus for cz"
    )


def test_allowed_move_qubits(sample_move_architecture):
    """
    Tests that instruction validation passes for allowed MOVE loci
    """
    arch = QuantumArchitecture(**sample_move_architecture).quantum_architecture
    IQMClient._validate_instruction(arch, Instruction(name='move', qubits=['QB3', 'COMP_R'], args={}), None)
    IQMClient._validate_instruction(
        arch, Instruction(name='move', qubits=['3', '0'], args={}), qubit_mapping={'3': 'QB3', '0': 'COMP_R'}
    )
    IQMClient._validate_instruction(
        arch, Instruction(name='move', qubits=['0', '1'], args={}), qubit_mapping={'0': 'QB3', '1': 'COMP_R'}
    )


@pytest.mark.parametrize(
    'qubits', [['QB1', 'QB2'], ['QB2', 'QB1'], ['QB1', 'QB1'], ['QB3', 'QB1'], ['COMP_R', 'COMP_R'], ['COMP_R', 'QB3']]
)
def test_disallowed_move_qubits(sample_move_architecture, qubits):
    """
    Tests that instruction validation fails for loci that are not allowed for MOVE by the quantum architecture
    """
    arch = QuantumArchitecture(**sample_move_architecture).quantum_architecture
    with pytest.raises(CircuitExecutionError) as err:
        IQMClient._validate_instruction(arch, Instruction(name='move', qubits=qubits, args={}), None)
    assert str(err.value) == f"('{qubits[0]}', '{qubits[1]}') not allowed as locus for move"


@pytest.mark.parametrize('qubits', [['QB1', 'QB2'], ['QB2'], ['QB1', 'QB2', 'QB3'], ['QB3', 'QB1'], ['QB1']])
def test_allowed_measure_qubits(sample_move_architecture, qubits):
    """
    Tests that instruction validation succeeds for loci that are any combination of valid measure qubits
    """
    arch = QuantumArchitecture(**sample_move_architecture).quantum_architecture
    IQMClient._validate_instruction(arch, Instruction(name='measure', qubits=qubits, args={'key': 'measure_1'}), None)


@pytest.mark.parametrize('qubits', [['QB1', 'COMP_R'], ['COMP_R'], ['QB1', 'QB2', 'QB4'], ['QB4']])
def test_disallowed_measure_qubits(sample_move_architecture, qubits):
    """
    Tests that instruction validation fails for loci containing any qubits that are not valid measure qubits
    """
    arch = QuantumArchitecture(**sample_move_architecture).quantum_architecture
    with pytest.raises(CircuitExecutionError) as err:
        IQMClient._validate_instruction(
            arch, Instruction(name='measure', qubits=qubits, args={'key': 'measure_1'}), None
        )
    assert re.match(r'Qubit (.+) is not allowed as locus for measure', str(err.value))


def test_barrier(sample_move_architecture):
    """
    Tests that instruction validation passes for the barrier operation
    """
    arch = QuantumArchitecture(**sample_move_architecture).quantum_architecture
    IQMClient._validate_instruction(
        arch, Instruction(name='barrier', qubits=['COMP_R', 'QB1', 'QB2', 'QB3'], args={}), None
    )
    IQMClient._validate_instruction(
        arch, Instruction(name='barrier', qubits=['QB1', 'COMP_R', 'QB2', 'QB3'], args={}), None
    )



# Copyright 2024 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
This module contains the definition of the Instruction class and defines
the supported instructions.

For more information on the topic, refer to the more comprehensive documentation in iqm_client.py.
"""

from typing import Any, Optional

from pydantic import BaseModel, Field, StrictStr, field_validator
from pydantic_core.core_schema import ValidationInfo

# Supported instruction configuration values:
# - 'arity': the arity of the locus; use -1 to allow any arity.
# - 'args': allowed arguments for the operation.
# - 'directed': if True, the loci defined in the architecture description are considered directed.
# - 'renamed_to': if set, indicates that this instruction name is deprecated, and IQM client will
#    auto-rename it to the new name.
# - 'check_locus': if False, skips the locus checking entirely. If set to the string
#   value 'any_combination', will check that the qubits in the instruction locus are found
#   in the architecture definition, but will allow any combination of them in.

SUPPORTED_INSTRUCTIONS: dict[str, dict[str, Any]] = {
    'barrier': {
        'arity': -1,
        'args': {},
        'check_locus': False,
    },
    'cz': {
        'arity': 2,
        'args': {},
        'directed': False,
    },
    'move': {
        'arity': 2,
        'args': {},
        'directed': True,
    },
    'measure': {
        'arity': -1,
        'args': {
            'key': (str,),
        },
        'check_locus': 'any_combination',
    },
    'measurement': {  # deprecated
        'arity': -1,
        'args': {
            'key': (str,),
        },
        'renamed_to': 'measure',
        'check_locus': 'any_combination',
    },
    'prx': {
        'arity': 1,
        'args': {
            'angle_t': (
                float,
                int,
            ),
            'phase_t': (
                float,
                int,
            ),
        },
    },
    'phased_rx': {  # deprecated
        'arity': 1,
        'args': {
            'angle_t': (
                float,
                int,
            ),
            'phase_t': (
                float,
                int,
            ),
        },
        'renamed_to': 'prx',
    },
}


class Instruction(BaseModel):
    """An instruction in a quantum circuit."""

    name: str = Field(..., examples=['measure'])
    """name of the quantum operation"""
    implementation: Optional[StrictStr] = Field(None)
    """name of the implementation, for experimental use only"""
    qubits: tuple[StrictStr, ...] = Field(..., examples=[('alice',)])
    """names of the logical qubits the operation acts on"""
    args: dict[str, Any] = Field(..., examples=[{'key': 'm'}])
    """arguments for the operation"""

    def __init__(self, **data):
        super().__init__(**data)
        # Auto-convert name if a deprecated name is used
        self.name = get_current_instruction_name(self.name)

    @field_validator('name')
    @classmethod
    def name_validator(cls, value):
        """Check if the name of instruction is set to one of the supported quantum operations"""
        name = value
        if name not in SUPPORTED_INSTRUCTIONS:
            raise ValueError(
                f'Unknown instruction "{name}". '
                f'Supported instructions are \"{", ".join(SUPPORTED_INSTRUCTIONS.keys())}\"'
            )
        return name

    @field_validator('implementation')
    @classmethod
    def implementation_validator(cls, value):
        """Check if the implementation of the instruction is set to a non-empty string"""
        implementation = value
        if isinstance(implementation, str):
            if not implementation:
                raise ValueError('Implementation of the instruction should be None, or a non-empty string')
        return implementation

    @field_validator('qubits')
    @classmethod
    def qubits_validator(cls, value, info: ValidationInfo):
        """Check if the instruction has the correct number of qubits according to the instruction's type"""
        qubits = value
        name = info.data.get('name')
        if not name:
            raise ValueError('Could not validate qubits because the name of the instruction did not pass validation')
        arity = SUPPORTED_INSTRUCTIONS[name]['arity']
        if (0 <= arity) and (arity != len(qubits)):
            raise ValueError(
                f'The "{name}" instruction acts on {arity} qubit(s), but {len(qubits)} were given: {qubits}'
            )
        return qubits

    @field_validator('args')
    @classmethod
    def args_validator(cls, value, info: ValidationInfo):
        """Check argument names and types for a given instruction"""
        args = value
        name = info.data.get('name')
        if not name:
            raise ValueError('Could not validate args because the name of the instruction did not pass validation')

        # Check argument names
        submitted_arg_names = set(args.keys())
        supported_arg_names = set(SUPPORTED_INSTRUCTIONS[name]['args'].keys())
        if submitted_arg_names != supported_arg_names:
            raise ValueError(
                f'The instruction "{name}" requires '
                f'{tuple(supported_arg_names) if supported_arg_names else "no"} argument(s), '
                f'but {tuple(submitted_arg_names)} were given'
            )

        # Check argument types
        for arg_name, arg_value in args.items():
            supported_arg_types = SUPPORTED_INSTRUCTIONS[name]['args'][arg_name]
            if not isinstance(arg_value, supported_arg_types):
                raise TypeError(
                    f'The argument "{arg_name}" should be of one of the following supported types'
                    f' {supported_arg_types}, but ({type(arg_value)}) was given'
                )

        return value


def is_multi_qubit_instruction(name: str) -> bool:
    """Checks if the instruction with the given name is a multi-qubit instruction.

    Args:
        name: The name of the instruction to check

    Returns:
        True if the instruction expects more than one qubit as its locus.
    """
    return name in SUPPORTED_INSTRUCTIONS and SUPPORTED_INSTRUCTIONS[name]['arity'] > 1


def is_directed_instruction(name: str) -> bool:
    """Checks if the instruction with the given name is directed, i.e. if the instruction
    is allowed only in the direction defined by the operation loci in the quantum architecture.

    Args:
        name: The name of the instruction to check

    Returns:
        True if the instruction is valid only in the direction defined by the architecture specification; or
        False if the instruction is valid also with the qubits in the locus in reversed order.
    """
    return (
        name in SUPPORTED_INSTRUCTIONS
        and 'directed' in SUPPORTED_INSTRUCTIONS[name]
        and SUPPORTED_INSTRUCTIONS[name]['directed']
    )


def get_current_instruction_name(name: str):
    """Checks if the instruction name has been deprecated and returns the new name if it is;
    otherwise, just returns the name as-is.

    Args:
        name: the name of the instruction

    Returns:
        the current name of the instruction.
    """
    return (
        SUPPORTED_INSTRUCTIONS[name]['renamed_to']
        if name in SUPPORTED_INSTRUCTIONS and 'renamed_to' in SUPPORTED_INSTRUCTIONS[name]
        else name
    )



# Copyright 2022 IQM client developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for user authentication and token management in IQM client.
"""
# pylint: disable=too-many-arguments
import builtins
import io
import json
from time import sleep

from mockito import expect, unstub, verifyNoUnwantedInteractions, when
import pytest
import requests

from iqm.iqm_client import ClientAuthenticationError, ClientConfigurationError, Credentials, IQMClient
from iqm.iqm_client.iqm_client import _time_left_seconds
from tests.conftest import (
    MockJsonResponse,
    expect_logout,
    get_jobs_args,
    make_token,
    post_jobs_args,
    prepare_tokens,
    submit_circuits_args,
)


def test_get_initial_tokens_with_credentials_from_arguments(credentials):
    """
    Tests that if the client is initialized with credentials, they are used correctly
    """
    tokens = prepare_tokens(300, 3600, **credentials)
    expected_credentials = Credentials(
        access_token=tokens['access_token'], refresh_token=tokens['refresh_token'], **credentials
    )

    client = IQMClient('https://example.com', **credentials)
    assert client._credentials == expected_credentials

    unstub()


def test_get_token_from_arguments(monkeypatch):
    """
    Tests that plaintext token is read from the arguments
    """
    monkeypatch.setenv('IQM_TOKEN', '2345')
    client = IQMClient(url='https://example.com', token='1234')
    assert client._token == '1234'


def test_get_initial_tokens_with_credentials_from_env_variables(credentials, monkeypatch):
    """
    Tests that credentials are read from environment variables if they are not given as arguments
    """
    tokens = prepare_tokens(300, 3600, **credentials)
    expected_credentials = Credentials(
        access_token=tokens['access_token'], refresh_token=tokens['refresh_token'], **credentials
    )
    monkeypatch.setenv('IQM_AUTH_SERVER', credentials['auth_server_url'])
    monkeypatch.setenv('IQM_AUTH_USERNAME', credentials['username'])
    monkeypatch.setenv('IQM_AUTH_PASSWORD', credentials['password'])

    client = IQMClient(url='https://example.com')
    assert client._credentials == expected_credentials
    unstub()


def test_get_token_from_env_variable(monkeypatch):
    """
    Tests that plaintext token is read from the environment variable if they are not given as arguments
    """
    monkeypatch.setenv('IQM_TOKEN', '1234')
    client = IQMClient(url='https://example.com')
    assert client._token == '1234'


def test_get_initial_tokens_with_incomplete_credentials_from_env_variables(credentials, monkeypatch):
    """
    Tests configuration error is reported if IQM_AUTH_SERVER is set, but no credentials provided
    """
    monkeypatch.setenv('IQM_AUTH_SERVER', credentials['auth_server_url'])
    with pytest.raises(ClientConfigurationError) as e:
        IQMClient('https://example.com')
    assert str(e.value) == 'Auth server URL is set but no username or password'
    unstub()


def test_add_authorization_header_when_credentials_are_provided(
    client_with_credentials, existing_job_url, existing_run_id, pending_compilation_job_result
):
    """
    Tests that ``get_run`` requests are sent with Authorization header when credentials are provided
    """
    expect(requests, times=1).get(
        existing_job_url, **post_jobs_args(access_token=client_with_credentials._credentials.access_token)
    ).thenReturn(pending_compilation_job_result)

    result = client_with_credentials.get_run(existing_run_id)
    assert result.status == 'pending compilation'

    verifyNoUnwantedInteractions()
    unstub()


def test_add_authorization_header_on_submit_circuits_when_credentials_are_provided(
    client_with_credentials,
    jobs_url,
    minimal_run_request,
    existing_run_id,
    submit_success,
    quantum_architecture_url,
    quantum_architecture_success,
):
    """
    Tests that ``submit_circuits`` requests are sent with Authorization header when credentials are provided
    """
    expect(requests, times=1).get(quantum_architecture_url, ...).thenReturn(quantum_architecture_success)
    expect(requests, times=1).post(
        jobs_url, **post_jobs_args(minimal_run_request, access_token=client_with_credentials._credentials.access_token)
    ).thenReturn(submit_success)

    assert existing_run_id == client_with_credentials.submit_circuits(**submit_circuits_args(minimal_run_request))

    verifyNoUnwantedInteractions()
    unstub()


def test_submit_circuits_raises_when_auth_failed(
    client_with_credentials,
    jobs_url,
    minimal_run_request,
    submit_failed_auth,
    quantum_architecture_url,
    quantum_architecture_success,
):
    """
    Tests that ``submit_circuits`` raises ClientAuthenticationError when authentication fails
    """
    expect(requests, times=1).get(quantum_architecture_url, ...).thenReturn(quantum_architecture_success)
    expect(requests, times=1).post(
        jobs_url, **post_jobs_args(minimal_run_request, access_token=client_with_credentials._credentials.access_token)
    ).thenReturn(submit_failed_auth)

    with pytest.raises(ClientAuthenticationError) as e:
        client_with_credentials.submit_circuits(**submit_circuits_args(minimal_run_request))
    assert str(e.value).startswith('Authentication failed')

    verifyNoUnwantedInteractions()
    unstub()


def test_add_authorization_header_on_get_jobs_when_external_token_is_provided(
    client_with_external_token, existing_job_url, tokens_dict, existing_run_id, pending_execution_job_result
):
    """
    Tests that get jobs requests are sent with Authorization header when external token is provided
    """
    expect(requests, times=1).get(
        existing_job_url, **get_jobs_args(access_token=tokens_dict['access_token'])
    ).thenReturn(pending_execution_job_result)

    result = client_with_external_token.get_run(existing_run_id)
    assert result.status == 'pending execution'

    verifyNoUnwantedInteractions()
    unstub()


def test_no_authorization_header_on_get_jobs_when_credentials_are_not_provided(
    sample_client, existing_job_url, existing_run_id, pending_execution_job_result
):
    """
    Tests that get jobs requests are sent without Authorization header when no credentials are provided
    """
    expect(requests, times=1).get(existing_job_url, **get_jobs_args(access_token=None)).thenReturn(
        pending_execution_job_result
    )

    result = sample_client.get_run(existing_run_id)
    assert result.status == 'pending execution'

    verifyNoUnwantedInteractions()
    unstub()


def test_raises_client_authentication_error_if_authentication_fails(base_url, credentials):
    """
    Tests that authentication failure raises ClientAuthenticationError
    """
    prepare_tokens(300, 3600, status_code=401, **credentials)
    with pytest.raises(ClientAuthenticationError) as e:
        IQMClient(base_url, **credentials)
    assert str(e.value).startswith('Failed to update tokens')

    verifyNoUnwantedInteractions()
    unstub()


def test_get_quantum_architecture_raises_if_no_auth_provided(sample_client, quantum_architecture_url):
    """Test retrieving the quantum architecture if server responded with redirect"""
    redirection_response = requests.Response()
    redirection_response.status_code = 302

    expect(requests, times=1).get(quantum_architecture_url, **get_jobs_args(access_token=None)).thenReturn(
        MockJsonResponse(401, {'detail': 'unauthorized'}, [redirection_response])
    )

    with pytest.raises(ClientConfigurationError) as e:
        sample_client.get_quantum_architecture()
    assert str(e.value) == 'Authentication is required.'

    verifyNoUnwantedInteractions()
    unstub()


def test_get_quantum_architecture_raises_if_wrong_auth_provided(client_with_credentials, quantum_architecture_url):
    """Test retrieving the quantum architecture if server responded with auth error"""
    expect(requests, times=1).get(
        quantum_architecture_url, **get_jobs_args(access_token=client_with_credentials._credentials.access_token)
    ).thenReturn(MockJsonResponse(401, {'detail': 'unauthorized'}))

    with pytest.raises(ClientAuthenticationError) as e:
        client_with_credentials.get_quantum_architecture()
    assert str(e.value).startswith('Authentication failed')

    verifyNoUnwantedInteractions()
    unstub()


def test_access_token_is_not_refreshed_if_it_has_not_expired(
    client_with_credentials, credentials, existing_job_url, existing_run_id, pending_execution_job_result
):
    """
    Test that access token is not refreshed if it has not expired
    """
    tokens = prepare_tokens(300, 3600, **credentials)
    assert client_with_credentials._credentials.access_token == tokens['access_token']

    expect(requests, times=3).get(
        existing_job_url, **get_jobs_args(access_token=client_with_credentials._credentials.access_token)
    ).thenReturn(pending_execution_job_result)

    client_with_credentials.get_run(existing_run_id)
    client_with_credentials.get_run(existing_run_id)
    client_with_credentials.get_run(existing_run_id)

    verifyNoUnwantedInteractions()
    unstub()


def test_expired_access_token_is_refreshed_automatically(
    client_with_credentials, credentials, existing_job_url, existing_run_id, pending_execution_job_result
):
    """
    Test that access token is refreshed automatically if it has expired
    """
    # Provide client with an expired access token
    client_with_credentials._credentials.access_token = make_token('Bearer', -300)
    client_with_credentials._credentials.refresh_token = make_token('Refresh', 4200)

    # Prepare for token refresh request
    refreshed_tokens = prepare_tokens(300, 4200, client_with_credentials._credentials.refresh_token, **credentials)

    # Expect get jobs request with refreshed token
    expect(requests, times=1).get(
        existing_job_url, **get_jobs_args(access_token=refreshed_tokens['access_token'])
    ).thenReturn(pending_execution_job_result)

    # Client should refresh tokens for get jobs request
    result = client_with_credentials.get_run(existing_run_id)
    assert result.status == 'pending execution'

    # Verify that access token has been refreshed
    assert client_with_credentials._credentials.access_token == refreshed_tokens['access_token']

    verifyNoUnwantedInteractions()
    unstub()


def test_start_new_session_when_refresh_token_has_expired(
    client_with_credentials, credentials, existing_job_url, existing_run_id, pending_execution_job_result
):
    """
    Test that a new session is started automatically if refresh token has expired
    """
    # Provide client with an expired tokens
    client_with_credentials._credentials.access_token = make_token('Bearer', -300)
    client_with_credentials._credentials.refresh_token = make_token('Refresh', -300)

    # Prepare for new session start instead of token refresh request
    refreshed_tokens = prepare_tokens(300, 3600, **credentials)

    # Expect get jobs request with refreshed token
    expect(requests, times=1).get(
        existing_job_url, **get_jobs_args(access_token=refreshed_tokens['access_token'])
    ).thenReturn(pending_execution_job_result)

    # Client should refresh tokens for get jobs request
    result = client_with_credentials.get_run(existing_run_id)
    assert result.status == 'pending execution'

    assert client_with_credentials._credentials.access_token == refreshed_tokens['access_token']
    assert client_with_credentials._credentials.refresh_token == refreshed_tokens['refresh_token']

    verifyNoUnwantedInteractions()
    unstub()


def test_tokens_are_cleared_at_logout(client_with_credentials, credentials):
    """
    Tests that calling ``close`` will terminate the session and clear tokens
    """
    expect_logout(credentials['auth_server_url'], client_with_credentials._credentials.refresh_token)

    client_with_credentials.close_auth_session()
    assert client_with_credentials._credentials.access_token is None
    assert client_with_credentials._credentials.refresh_token is None

    verifyNoUnwantedInteractions()
    unstub()


def test_cannot_close_external_auth_session(client_with_external_token):
    """
    Tests that calling ``close_auth_session`` while initialized with an external auth session
    raises ClientAuthenticationError
    """
    with pytest.raises(ClientAuthenticationError) as e:
        client_with_external_token.close_auth_session()
    assert 'Unable to close externally managed auth session' == str(e.value)


def test_logout_on_client_destruction(base_url, credentials):
    """
    Tests that client is trying to terminate the authentication session on destruction
    """
    prepare_tokens(300, 300, **credentials)
    client = IQMClient(base_url, **credentials)

    expect_logout(credentials['auth_server_url'], client._credentials.refresh_token)
    del client

    verifyNoUnwantedInteractions()
    unstub()


def test_external_token_updated_if_expired(base_url):
    """
    Tests that client gets updated token from tokens file if old external token has expired
    """
    tokens_path = 'dummy_path'

    def setup_tokens_file(access_token_lifetime):
        access_token = make_token('Bearer', access_token_lifetime)
        tokens_file_contents = json.dumps({'auth_server_url': base_url + '/auth', 'access_token': access_token})
        when(builtins).open(tokens_path, 'r', encoding='utf-8').thenReturn(io.StringIO(tokens_file_contents))

    setup_tokens_file(1)
    client = IQMClient(base_url, tokens_file=tokens_path)
    sleep(2)
    setup_tokens_file(300)
    bearer_token = client._get_bearer_token()
    assert bearer_token is not None
    assert _time_left_seconds(bearer_token.removeprefix('Bearer ')) > 0
    unstub()


def test_tokens_file_overrides_plaintext_token(base_url):
    """
    Tests tokens_file is preferred over plaintext token
    """
    tokens_path = 'dummy_path'
    access_token = make_token('Bearer', lifetime=300)
    tokens_file_contents = json.dumps({'auth_server_url': base_url + '/auth', 'access_token': access_token})
    when(builtins).open(tokens_path, 'r', encoding='utf-8').thenReturn(io.StringIO(tokens_file_contents))
    try:
        client = IQMClient(base_url, tokens_file=tokens_path, token='12345')
        bearer_token = client._get_bearer_token()
        assert bearer_token is not None
        assert bearer_token == f'Bearer {access_token}'
    finally:
        unstub()


def test_plaintext_token_is_used_as_bearer_token_if_tokens_file_is_not_set(base_url):
    """
    Tests that plaintext token is used as a bearer token correctly
    """
    client = IQMClient(base_url, token='1234')
    bearer_token = client._get_bearer_token()
    assert bearer_token == 'Bearer 1234'
